"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.UserRepository = void 0;
const tslib_1 = require("tslib");
// Copyright (c) 2023 Sourcefuse Technologies
//
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT
const core_1 = require("@loopback/core");
const repository_1 = require("@loopback/repository");
const rest_1 = require("@loopback/rest");
const core_2 = require("@sourceloop/core");
const keys_1 = require("../keys");
const models_1 = require("../models");
const types_1 = require("../types");
const otp_repository_1 = require("./otp.repository");
const user_credentials_repository_1 = require("./user-credentials.repository");
let UserRepository = class UserRepository extends core_2.DefaultSoftCrudRepository {
    constructor(dataSource, getUserCredsRepository, getOtpRepository, tenantRepositoryGetter, userTenantRepositoryGetter, logger, passwordDecryptionFn, passwordHashingFn, passwordVerifyFn, user) {
        super(user, dataSource);
        this.getOtpRepository = getOtpRepository;
        this.tenantRepositoryGetter = tenantRepositoryGetter;
        this.userTenantRepositoryGetter = userTenantRepositoryGetter;
        this.logger = logger;
        this.passwordDecryptionFn = passwordDecryptionFn;
        this.passwordHashingFn = passwordHashingFn;
        this.passwordVerifyFn = passwordVerifyFn;
        this.user = user;
        this.userTenants = this.createHasManyRepositoryFactoryFor('userTenants', userTenantRepositoryGetter);
        this.registerInclusionResolver('userTenants', this.userTenants.inclusionResolver);
        this.tenant = this.createBelongsToAccessorFor('defaultTenant', tenantRepositoryGetter);
        this.registerInclusionResolver('defaultTenant', this.tenant.inclusionResolver);
        this.credentials = this.createHasOneRepositoryFactoryFor('credentials', getUserCredsRepository);
        this.registerInclusionResolver('credentials', this.credentials.inclusionResolver);
    }
    async create(entity, options) {
        const user = await super.create(entity, options);
        try {
            // Add temporary password for first time
            const password = await this.passwordHashingFn();
            const creds = new models_1.UserCredentials({
                authProvider: 'internal',
                password,
            });
            await this.credentials(user.id).create(creds);
        }
        catch (err) {
            throw new rest_1.HttpErrors.UnprocessableEntity('Error while hashing password');
        }
        return user;
    }
    async createWithoutPassword(entity, options) {
        return super.create(entity, options);
    }
    async verifyPassword(username, password) {
        let newPassword = password;
        if (process.env.PRIVATE_DECRYPTION_KEY) {
            const decryptedPassword = await this.passwordDecryptionFn(password);
            newPassword = decryptedPassword;
        }
        const user = await super.findOne({
            where: { username: username.toLowerCase() },
        });
        const creds = user && (await this.credentials(user.id).get());
        if (!user || user.deleted) {
            throw new rest_1.HttpErrors.Unauthorized("UserDoesNotExist" /* AuthenticateErrorKeys.UserDoesNotExist */);
        }
        else if (!(creds === null || creds === void 0 ? void 0 : creds.password) ||
            creds.authProvider !== core_2.AuthProvider.INTERNAL ||
            !(await this.passwordVerifyFn(newPassword, creds.password))) {
            this.logger.error('User creds not found in DB or is invalid');
            throw new rest_1.HttpErrors.Unauthorized("Invalid Credentials" /* AuthErrorKeys.InvalidCredentials */);
        }
        else {
            return user;
        }
    }
    async decryptPassword(password) {
        const decryptedPassword = await this.passwordDecryptionFn(password);
        return decryptedPassword;
    }
    async updatePassword(username, password, newPassword) {
        var _a;
        const user = await super.findOne({ where: { username } });
        const creds = user && (await this.credentials(user.id).get());
        if ((_a = (!user || user.deleted)) !== null && _a !== void 0 ? _a : !(creds === null || creds === void 0 ? void 0 : creds.password)) {
            throw new rest_1.HttpErrors.Unauthorized("UserDoesNotExist" /* AuthenticateErrorKeys.UserDoesNotExist */);
        }
        else if (creds.authProvider !== core_2.AuthProvider.INTERNAL) {
            throw new rest_1.HttpErrors.BadRequest("PasswordCannotBeChangedForExternalUser" /* AuthenticateErrorKeys.PasswordCannotBeChanged */);
        }
        else if (!(await this.passwordVerifyFn(password, creds.password))) {
            throw new rest_1.HttpErrors.Unauthorized("Incorrect Password" /* AuthErrorKeys.WrongPassword */);
        }
        else if (await this.passwordVerifyFn(newPassword, creds.password)) {
            throw new rest_1.HttpErrors.Unauthorized('Password cannot be same as previous password!');
        }
        else {
            // Do nothing
        }
        await this.credentials(user.id).patch({
            password: await this.passwordHashingFn(newPassword),
        });
        return user;
    }
    async changePassword(username, newPassword, oldPassword) {
        var _a;
        const user = await super.findOne({ where: { username } });
        const creds = user && (await this.credentials(user.id).get());
        if (oldPassword) {
            // This method considers old password as OTP
            const otp = await (await this.getOtpRepository()).get(username);
            if (!otp || otp.otp !== oldPassword) {
                throw new rest_1.HttpErrors.Unauthorized("Incorrect Password" /* AuthErrorKeys.WrongPassword */);
            }
        }
        if ((creds === null || creds === void 0 ? void 0 : creds.authProvider) !== core_2.AuthProvider.INTERNAL) {
            throw new rest_1.HttpErrors.Unauthorized("PasswordCannotBeChangedForExternalUser" /* AuthenticateErrorKeys.PasswordCannotBeChanged */);
        }
        if ((_a = (!user || user.deleted)) !== null && _a !== void 0 ? _a : !(creds === null || creds === void 0 ? void 0 : creds.password)) {
            throw new rest_1.HttpErrors.Unauthorized("UserDoesNotExist" /* AuthenticateErrorKeys.UserDoesNotExist */);
        }
        else if (await this.passwordVerifyFn(newPassword, creds.password)) {
            throw new rest_1.HttpErrors.Unauthorized('Password cannot be same as previous password!');
        }
        else {
            // DO nothing
        }
        await this.credentials(user.id).patch({
            password: await this.passwordHashingFn(newPassword),
        });
        return user;
    }
    async updateLastLogin(userId) {
        await super.updateById(userId, {
            lastLogin: Date.now(),
        }, {
            currentUser: { id: userId },
        });
    }
    async firstTimeUser(userId) {
        const user = await super.findOne({
            where: {
                id: userId,
            },
        });
        if (!user) {
            throw new rest_1.HttpErrors.NotFound("UserDoesNotExist" /* AuthenticateErrorKeys.UserDoesNotExist */);
        }
        const userTenant = await (await this.userTenantRepositoryGetter()).findOne({
            where: {
                userId,
                tenantId: user.defaultTenantId,
                status: {
                    inq: [0 /* UserStatus.REGISTERED */, 3 /* UserStatus.PASSWORD_CHANGE_NEEDED */],
                },
            },
        });
        return !!userTenant;
    }
};
exports.UserRepository = UserRepository;
exports.UserRepository = UserRepository = tslib_1.__decorate([
    tslib_1.__param(0, (0, core_1.inject)(`datasources.${types_1.AuthDbSourceName}`)),
    tslib_1.__param(1, repository_1.repository.getter(user_credentials_repository_1.UserCredentialsRepository)),
    tslib_1.__param(2, repository_1.repository.getter(otp_repository_1.OtpRepository)),
    tslib_1.__param(3, repository_1.repository.getter('TenantRepository')),
    tslib_1.__param(4, repository_1.repository.getter('UserTenantRepository')),
    tslib_1.__param(5, (0, core_1.inject)(core_2.LOGGER.LOGGER_INJECT)),
    tslib_1.__param(6, (0, core_1.inject)(keys_1.AuthServiceBindings.PASSWORD_DECRYPTION_PROVIDER)),
    tslib_1.__param(7, (0, core_1.inject)(keys_1.AuthServiceBindings.PASSWORD_HASHING_PROVIDER)),
    tslib_1.__param(8, (0, core_1.inject)(keys_1.AuthServiceBindings.PASSWORD_VERIFY_PROVIDER)),
    tslib_1.__param(9, (0, core_1.inject)('models.User')),
    tslib_1.__metadata("design:paramtypes", [repository_1.juggler.DataSource, Function, Function, Function, Function, Object, Function, Function, Function, Object])
], UserRepository);
//# sourceMappingURL=user.repository.js.map